/*
 *	zx2gif
 *
 *	Copyright © 2002–2022 Jan Bobrowski
 *
 *	This program is free software; you can redistribute it and/or
 *	modify it under the terms of the GNU General Public License
 *	version 2 as published by the Free Software Foundation.
 */

#include <stdint.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <err.h>

#if defined __GLIBC__ || defined __BSD__
FILE *fopenz(const char *name, const char *mode);
#define FOPEN fopenz
#else
#define FOPEN fopen
#endif

#define STR_(T) #T
#define STR(T) STR_(T)
#define elemof(T) (sizeof T/sizeof*T)
#define endof(T) (T+elemof(T))
typedef uint8_t u8;
typedef uint16_t u16;
typedef uint32_t u32;

struct rect {
	int x, y, w, h;
};

struct img {
	int w, h;
	u8 *buf, *ptr, *end;
};

static u8 scr[6912];
static u8 border;

static FILE *in, *out;

static u8 get8()
{
	int v = getc(in);
	if (v == EOF)
		errx(1, "Premature end of file");
	return v;
}

static u16 get16()
{
	u16 v = get8();
	return v | get8()<<8;
}

static void skip(int n)
{
	while (--n >= 0)
		get8();
}

static void load_scr()
{
	int l = fread(scr, 1, sizeof scr, in);
	if (l < sizeof scr) errx(1, "Premature end of file");
}

static void load_tap()
{
	for (;;) {
		int l = get16();
		if (l - 2 < sizeof scr) {
			skip(l);
			continue;
		}
		get8();
		load_scr();
		return;
	}
}

static void load_sna()
{
	skip(26);
	border = get8() & 7;
	load_scr();
}

static void load_z80()
{
	u8 v;
	int oldfmt;
	skip(6); //0+6
	oldfmt = get8() | get8();
	skip(4); //8+4
	v = get8();
	if (v == 0xFF)
		v = 1;
	border = v>>1 & 7;
	skip(17); //13+17

	if (oldfmt) {
		if(v & 0x20)
			goto load_compr;
		load_scr();
		return;
	}
	{
		int cnt = get16();
		int pg, mode, p7FFD;
		if (cnt < 6) errx(1, "Z80: Truncated");
		skip(2);
		mode = get8();
		p7FFD = get8();
		if (mode <= 2 || cnt > 23 && mode <= 3 || mode == 128)
			p7FFD = 0;
		pg = p7FFD & 8 ? 10 : 8;
		cnt -= 4;
		for (;;) {
			skip(cnt);
			cnt = get16();
			if (get8() == pg)
				break;
		}
	}

load_compr:
	{ /* compressed */
		u8 *p = scr;
		u8 b = get8();
		do {
			u8 b2 = get8();
			if (b != 0xED || b2 != 0xED) {
				*p++ = b;
				b = b2;
				continue;
			}
			b = get8();
			b2 = get8();
			if (b > endof(scr) - p)
				b = endof(scr) - p;
			memset(p, b2, b); p += b;
			b = get8();
		} while (p < endof(scr));
	}
}

static void put8(u8 b) { fputc(b, out); }
static void put16(u16 s) { fputc(s & 0xFF, out); fputc(s>>8 & 0xFF, out); }

#define TRANSP 64
static int used[256];

static void round_img(struct img *i, int m)
{
	int hsz = i->w < 5 ? i->w : 5;
	for (int n = 0; n < 4; n++) if(m & 1<<n) {
		u8 *p = i->buf;
		int s = i->w;
		if (n & 1)
			p += s - hsz;
		if (n & 2) {
			p += (i->h - 1)*s;
			s = -s;
		}
		memset(p, TRANSP, hsz); p += s + (n & 1 ? hsz - 2 : 0);
		memset(p, TRANSP, 2); p += s + (n & 1 ? 2-1 : 0);
		p[0] = p[s] = TRANSP;
		used[TRANSP]++;
	}
}

static u8 color_index(int n, int b)
{
	n |= n<<3 | b;
	if (n == 64)
		n = 0;
	return n;
}

static void putcell(u8 *d, int pitch, int x, int y, int inv)
{
	int s = x | y<<5 & 0xE0 | y<<8 & 0x1800;
	u8 attr = scr[6144 | y<<5 | x];
	u8 ink = color_index(attr & 7, attr & 0x40);
	u8 paper = color_index(attr>>3 & 7, attr & 0x40);
	u8 xor = 0;
	if (inv) {
		if (!(attr & 0x80)) {
			used[TRANSP]++;
			return;
		}
		xor = 0xFF;
	}
	for (y = 0; y < 8; y++) {
		int b = (scr[s] ^ xor)<<1 | 1;
		do {
			u8 c = b & 0x100 ? ink : paper;
			used[c]++;
			*d++ = c;
			b <<= 1;
		} while (b & 0xFF);
		s += 256; d += pitch - 8;
	}
}

static void putcell2(u8 *d, int pitch, int x, int y, int inv)
{
	int s = x | y<<5 & 0xE0 | y<<8 & 0x1800;
	u8 attr = scr[6144 | y<<5 | x];
	u8 ink = attr & 7;
	u8 paper = attr>>3 & 0x7;
	u8 xor = 0;
	u8 bright = attr & 0x40;
	u8 cn[5];
	if (inv) {
		if (!(attr & 0x80)) {
			used[TRANSP]++;
			return;
		}
		xor = 0xFF;
	}

	cn[0] = color_index(paper, bright);
	if (ink == paper)
		cn[1] = cn[2] = cn[3] = cn[4] = cn[0];
	else {
		cn[1] = attr & 0x7F;
		cn[3] = bright | ink<<3 | paper;
		cn[2] = (ink > paper ? cn[1] : cn[3]) | 128;
		cn[4] = color_index(ink, bright);
	}

	for (y = 0; y < 8; y += 2) {
		static u8 cnt[] = {0, 1, 1, 2};
		int b1 = scr[s] ^ xor;
		int b2 = scr[s + 256] ^ xor;
		for (x = 6; x >= 0; x -= 2) {
			int n = cnt[b1>>x & 3] + cnt[b2>>x & 3];
			u8 c = cn[n];
			*d++ = c;
			used[c]++;
		}
		s += 512; d += pitch - 4;
	}
}

static struct img *img;

static int getpix()
{
	if (img->ptr < img->end)
		return *img->ptr++;
	return EOF;
}

static void init_image(struct img *img, int w, int h, int bg)
{
	img->buf = malloc(w*h);
	memset(img->buf, bg, w*h);
	img->w = w; img->h = h;
	img->ptr = img->buf;
	img->end = img->buf + w*h;
}

static int cellsz = 8;

static void put_zx(struct img *img, int xx, int yy, struct rect *frag, int inv)
{
	u8 *d = img->buf + xx + img->w*yy;
	for (int y = 0; y < frag->h; y++) {
		for (int x = 0; x < frag->w; x++) {
			if (cellsz == 8)
				putcell(d, img->w, frag->x + x, frag->y + y, inv);
			else
				putcell2(d, img->w, frag->x + x, frag->y + y, inv);
			d += cellsz;
		}
		d += cellsz*(img->w - frag->w);
	}
}

void chkflash(struct rect *r)
{
	int fx1 = 32, fy1 = 24, fx2 = 0, fy2 = 0;
	u8 *a = scr + 6144;
	for (int y = 0; y < 24; y++)
		for (int x = 0; x < 32; x++)
			if (*a++ & 0x80) {
				if (fx1 > x) fx1 = x;
				if (fx2 < x) fx2 = x;
				if (fy1 > y) fy1 = y;
				fy2 = y;
			}
	r->x = fx1;
	r->y = fy1;
	r->w = fx2 - fx1 + 1;
	r->h = fy2 - fy1 + 1;
}

static u8 pal[768];

static int genpal()
{
	static const u8 color[] = {
		0x00, 0x00, 0x00, 0x00, 0x79, 0x9C, 0x8B, 0xB4,
		0xB8, 0x9C, 0xD5, 0xB4, 0xDD, 0xDD, 0xFF, 0xFF,
	};
	int n, i = 0;
	for (n = 0; n < elemof(used); n++) {
		int m;
		if (!used[n])
			continue;
		used[n] = i;

		m = (n & 0x40 ? 2 : 0) | (n >= 128);
		pal[3*i+2] = color[m + (n & 8 | n<<2 & 4)];
		pal[3*i+0] = color[m + (n>>1 & 8 | n<<1 & 4)];
		pal[3*i+1] = color[m + (n>>2 & 8 | n & 4)];

		i++;
	}
	int b;
	for (b = 2; b <= 8; b++) {
		n = 1 << b;
		if (i <= n)
			break;
	}
	if (i < n)
		memset(pal + 3*i, 0, (n - i)*3);
	return b;
}

static void reindex(struct img *img)
{
	for (u8 *p = img->buf; p < img->end; p++)
		*p = used[*p];
}

static void putgce(int transp, int delay)
{
	put8('!'); // graphic control extension
	put8(0xF9);
	put8(4);
	put8(transp >= 0); // disposal|user inp.|transp fl.
	put16(delay); // delay [1/100s]
	put8(transp); // transparent color
	put8(0); // terminator
}

int compress(int init_bits, FILE* outfile, int (*readValue)());

int main(int argc, char *argv[])
{
	int ifmt = '$';
	struct rect flash;
	int borderw = 0;
	int borderc = -1;
	int offset = 0;
	int extract = 0;
	int round = 0;
	struct img mimg, fimg;
	int bpp;
	int o, tu = 0;
	char *name = 0;

	static char usage[] = 
		 "usage: zx2gif [-$TSZ] [-rh] [-b size] [-c color] [-n text] [-s offs] in out.gif\n"
		 "       zx2gif [-$TSZ] -x in out.scr\n";

	if (argc < 2) {
		fprintf(stderr, "%s       zx2gif -h\n", usage);
		return 1;
	}

	while ((o = getopt(argc, argv, "b:c:hxs:n:r$TSZ?")) != EOF) switch(o) {
		case 'Z':
		case 'S':
		case '$':
		case 'T': ifmt = o; break;
		case 'b': borderw = atoi(optarg); break;
		case 'c': borderc = atoi(optarg); borderc &= 0xF; break;
		case 'h': cellsz = 4; break;
		case 's': offset = atoi(optarg); break;
		case 'x': extract = 1; break;
		case 'n': name = optarg; break;
		case 'r': round = 1; break;
		case '?':
		default:
usage:
			printf(
			 "zx2gif "STR(VERSION)" by Jan Bobrowski; GPL\n"
			 "%s"
			 " -$ screen.scr (default)\t-h\thalf size\n"
			 " -T tape.tap\t\t\t-r\tround image\n"
			 " -S snapshot.sna\t\t-b num\tborder size\n"
			 " -Z snapshot.z80\t\t-c num\tborder colour\n"
			 " -n text\tcomment\n"
			 " -s offset\tposition in input file\n", usage);
			return 0;
	}

	if (argc > optind + 2)
		errx(1, "Too many arguments");

	in = stdin;
	out = stdout;

	if (optind < argc) {
		if (strcmp(argv[optind], "-")) {
			in = FOPEN(argv[optind], "rb");
			if (!in) err(1, "%s", argv[optind]);
		}
	} else if (isatty(0))
		goto usage;

	if (offset > 0)
		skip(offset);

	switch (ifmt) {
		case '$': load_scr(); break;
		case 'T': load_tap(); break;
		case 'S': load_sna(); break;
		case 'Z': load_z80(); break;
	}

	if (++optind < argc && strcmp(argv[optind], "-") != 0) {
		out = fopen(argv[optind], "wb");
		if (!out) err(1, "%s", argv[optind]);
	}

	if (extract) {
		fwrite(scr, 1, sizeof scr, out);
		return 0;
	}

	if (borderw) {
		if (borderc >= 0)
			border = borderc;
		border = color_index(border, 0);
		used[border]++;
	}

	chkflash(&flash);

	init_image(&mimg, 32*cellsz + 2*borderw, 24*cellsz + 2*borderw, border);
	{
		static struct rect full = {0,0,32,24};
		put_zx(&mimg, borderw, borderw, &full, 0);
		if (round)
			round_img(&mimg, 0b1111);
	}
	if (flash.w > 0) {
		init_image(&fimg, flash.w*cellsz, flash.h*cellsz, TRANSP);
		put_zx(&fimg, 0, 0, &flash, 1);
		if (round)
			round_img(&fimg, (flash.x ? 0b1010 : 15) & (flash.y ? 0b1100 : 15)
			 & (flash.x + flash.w < 32 ? 0b0101 : 15) & (flash.y + flash.h < 24 ? 0b0011 : 15));
	}

	tu = used[TRANSP];
	bpp = genpal();
	tu = tu > 0 ? used[TRANSP] : -1;

	reindex(&mimg);
	if (flash.w > 0)
		reindex(&fimg);

	fwrite(flash.w < 0 ? "GIF87a" : "GIF89a", 6, 1, out);
	put16(mimg.w); put16(mimg.h);
	put8(0xF0 | bpp-1);
//	put8(0x80 | bpp-1<<4 | bpp-1); ???
	put8(used[TRANSP]); //bg XXX
	put8(0); // aspect

	{
		int i, n = 3 << bpp;
		for (i = 0; i < n; i++)
			put8(pal[i]);
	}

	if (flash.w > 0) {
		put8(0x21); put8(0xFF);
		put8(11);
		fwrite("NETSCAPE2.0", 11, 1, out);
		put8(3);
		put8(1);
		put16(0);
		put8(0);

		putgce(round ? tu : -1, 32);
	} else if (round)
		putgce(tu, 0);

	put8(',');
	put16(0); put16(0);
	put16(mimg.w); put16(mimg.h);
	put8(0);
	img = &mimg;
	compress(bpp, out, getpix);

	if (flash.w > 0) {
		int x = flash.x*cellsz + borderw;
		int y = flash.y*cellsz + borderw;

		putgce(tu, 32);

		put8(',');
		put16(x); put16(y);
		put16(fimg.w); put16(fimg.h);
		put8(0);

		img = &fimg;
		compress(bpp, out, getpix);
	}

	if (name && *name) {
		// Insert Comment
		int l = strlen(name);
		if (l > 255)
			l = 255;
		put8('!'); put8(0xFE); put8(l);
		fwrite(name, 1, l, out);
		put8(0);
	}

	put8(';');

	return 0;
}
